/*
 * check_sap
 *
 * This program is designed to interface between SAP CCMS and Nagios
 * using rfcsdk from SAP
 *
 * William Preston, NETWAYS GmbH 2010
 * GPLv2 (see included LICENCE file for full copyright details)
 * 
 * check_sap -d <destination e.g. SAPID> -c <client e.g. 000> -u <user> -p <pass> [-l <lang>] [-3|-2 (version)] [-t (trace on)] \
 *  -h <hostname> -s <sysnr e.g. 00> [-g <gateway (if not hostname)>] [-x TCP/IP service default sapgwNN (where NN is sysnr)]
 *  [-m <modename>]
 *
 */

#include <unistd.h>
#include <getopt.h>
#include <string.h>
#include <stdio.h>
#include <regex.h>
#include <inttypes.h>
#include "mt.h"



/*
 * NOTE: Linux uses 4 Byte WChars, whereas SAP use 2 Byte WChars (UTF-16)
 * To work around this compile with -DWCHAR_is_2B -fshort-wchar
 * which will however break functions from libc like strlen...
 * Hacks to handle this are scattered through the code.
 * There must be a better way to do this??

    wchar_t test[] = cU("RF");
    size_t  fix;
    fix = sizeof(wchar_t)/strlenU(test);
*/


#define SAPwithUNICODE

#ifdef WCHAR_WORKAROUND
#undef WCHAR_is_4B
#define WCHAR_is_2B
#endif /* WCHAR_WORKAROUND */

#include "saprfc.h"
#include "sapitab.h"
#include "rfcsi.h"
#undef strlenU
#define strlenU(x) mystrlen(x)
#include "check_sap.h"


#define STRCPY(x,y) memcpy(x, y, (sizeofR(y)<sizeofR(x))?sizeofR(y):sizeofR(x))
#define STRCPYSTR(x,y) memcpy(x, y, (strlen(y)<sizeofR(x))?strlen(y):sizeofR(x))
#define STRCPYUSTR(x,y) memcpy(x, y, (strlenU(y)<sizeofR(x))?strlenU(y):sizeofR(x))
#define STRCPY2(x,y) memcpy(&x, y, sizeof(x))
#define STRADD(x,y) strncat(x, y, sizeof(x) - strlen(x) - 1)
#define FREE(x) if (x) { free(x); x = 0; }

/* flag to build fullpaths for the matching */
int flag_fullpath = 0;
#define PATH_SEPARATOR '/'


/* store the connection options in a struct */
struct sap_options_s
{
    char *client;
    char *dest;
    char *gw;
    char *host;
    char *sysnr;
    char *lang;
    char *user;
    char *pass;
    char *svc;
    int trace;
    int ver;
} sap_options;


/* enable tree caching? */
int cache = 0;


const char *modes[] = { "listsets", "listmonitors", "sapinfo", "listtree", NULL };
enum mode_enum { LISTSETS, LISTMONITORS, SAPINFO, LISTTREE, CHECK};

const char *exitval[] = { "UNKNOWN", "OK", "WARNING", "CRITICAL", NULL };

/* prototypes */
int build_param(char **, struct sap_options_s *);
char *quote_param(char *);
size_t mystrlen(const wchar_t *);
void str_param(RFC_PARAMETER *, SAP_UC *, SAP_UC *);
int xmi_getmslist(RFC_HANDLE, rfc_char_t *);
int xmi_getmonlist(RFC_HANDLE, rfc_char_t *, rfc_char_t *);
int xmi_getmontree(RFC_HANDLE, rfc_char_t *, rfc_char_t *, rfc_char_t *, BAPITNDEXT **);
#ifdef CACHE
int tree_from_cache(rfc_char_t *, rfc_char_t *, int, BAPITNDEXT **);
int tree_to_cache(rfc_char_t *, rcf_char_t *, int, BAPITNDEXT **);
#endif /* CACHE */
int displaytree(int, BAPITNDEXT *);
// int matchitems(int, BAPITNDEXT *, BAPITID ***, wchar_t *);
int matchitems(int, BAPITNDEXT *, MTE **, wchar_t *);
int mbsToUCS2(wchar_t **, const char *, size_t);
size_t mystrfill(wchar_t *, const wchar_t, size_t);
wchar_t *mystrclean(wchar_t *, size_t);
int xmi_get_value(RFC_HANDLE, rfc_char_t *, BAPITID, BAPIALDATA *);
int usage(char *);
void strcpy_nospaces(char *, char *);


int main(int argc, char **argv)
{
    char            *connstr;
    int             opt;
    char            *myname;

    RFC_ERROR_INFO_EX   error_info;
    RFC_HANDLE      handle = RFC_HANDLE_NULL;
    RFC_RC          rfc_rc = RFC_OK;

    BAPITNDEXT      *tree = 0;
    BAPITID         **tids = 0;
    BAPITID         current;
    BAPIPACVAL      perfval;
    BAPIPAPROP      perfprop;
    BAPISMCVAL      status;
    BAPIALDATA      data;

    int             i, tree_items, count, exitcode = 1;
    int             status_count[4];

    rfc_char_t      *buffer;
    rfc_char_t      *tmpstr;

    MTE             *mteptr; /* list of MTEs */
    MTE             *mteptr2;
    char            perfdata[MAXPERFDATA];
    char            output[MAXFIRSTLINE];
    char            multi_output[MAXMULTILINE];
    int             output_lines = 0;
    char            *valunit;
    char            *valstr;
    char            tmpbuf[1024];
    wchar_t         *cleanstr;

    /* thresholds returned by SAP */
    int             warn_threshold = 0;
    int             crit_threshold = 0;




    /* user supplied parameters */
    int             mode = CHECK;
    int             flag_thresholds = 0, flag_ranges = 0;
    rfc_char_t      *user;
    rfc_char_t      *monitor_set, *monitor_name;
    rfc_char_t      *match;

    /* fill options with defaults */
    sap_options.trace = 0;
    sap_options.svc = NULL;
    sap_options.ver = 3;
    sap_options.lang = "E";
    sap_options.client = "000";
    sap_options.sysnr = "00";
    sap_options.user = "user";
    sap_options.pass = "pass";
    sap_options.host = "localhost";
    sap_options.gw = "localhost";
    sap_options.dest = "SAP";

    memset(perfdata, 0, sizeof(perfdata));
    memset(output, 0, sizeof(output));
    memset(multi_output, 0, sizeof(multi_output));
    memset(status_count, 0, sizeof(int *)*sizeof(status_count));

    myname = argv[0];

    static struct option longopts[] =
    {
        {"client",      required_argument,  NULL,   'c'},
        {"dest",        required_argument,  NULL,   'd'},
        {"fullpath",    no_argument,        NULL,   'f'},
        {"gateway",     required_argument,  NULL,   'g'},
        {"host",        required_argument,  NULL,   'h'},
        {"lang",        required_argument,  NULL,   'l'},
        {"mode",        required_argument,  NULL,   'm'},
        {"pass",        required_argument,  NULL,   'p'},
        {"sysnr",       required_argument,  NULL,   's'},
        {"trace",       no_argument,        NULL,   't'},
        {"user",        required_argument,  NULL,   'u'},
        {"gwserv",      required_argument,  NULL,   'x'},
        {"help",        no_argument,        NULL,   '?'},
        {"thresholds",  no_argument,        NULL,   'j'},
        {"use-ranges",  no_argument,        NULL,   'k'},
#ifdef CACHE
        {"cache",       required_argument,  NULL,   'C'},
#endif /* CACHE */
        {NULL,          0,                  NULL,   0}
    };

    while ((opt = getopt_long(argc, argv, "c:d:fg:h:jl:m:p:s:tu:x:23C:", longopts, NULL)) != -1)
    {
        switch(opt)
        {
            case 'c':
                sap_options.client = optarg;
                break;
            case 'd':
                sap_options.dest = optarg;
                break;
            case 'f':
                flag_fullpath = 1;
                break;
            case 'g':
                sap_options.gw = optarg;
                break;
            case 'h':
                sap_options.host = optarg;
                break;
            case 'k':
                flag_ranges = 1;
                /* fall through to j */
            case 'j':
                flag_thresholds = 1;
                break;
            case 'l':
                sap_options.lang = optarg;
                break;
            case 'm':
                /* mode switch */
                for (i=0; modes[i]; i++)
                {
                    if (!strcmp(optarg, modes[i]))
                    {
                        mode = i;
                        break;
                    }
                }
                break;
            case 'p':
                sap_options.pass = optarg;
                break;
            case 's':
                sap_options.sysnr = optarg;
                break;
            case 't':
                sap_options.trace = 1;
                break;
            case 'u':
                sap_options.user = optarg;
                break;
            case 'x':
                sap_options.svc = optarg;
                break;
            case '2':
                sap_options.ver = 2;
                break;
            case 'C':
#ifdef CACHE
                cache = (int)strtol(optarg, NULL, 10);
#endif /* CACHE */
                break;
            case 0:
                /* long options */
                break;
            default:
                /* usage */
                return(usage(myname));
                break;
        }
    }
    argc -= optind;
    argv += optind;

    switch(mode)
    {
        case LISTSETS:
            if (argc != 0)
            {
                printf("Mode %s does not take parameters\n", modes[LISTSETS]);
                return(usage(myname));
            }
            break;
        case LISTMONITORS:
            if (argc != 1)
            {
                printf("Mode %s takes exactly 1 parameter\n", modes[LISTMONITORS]);
                return(usage(myname));
            }
            mbsToUCS2(&monitor_set, argv[0], strlen(argv[0]));
            break;
        case LISTTREE:
            if (argc != 2)
            {
                printf("Mode %s takes exactly 2 parameters\n", modes[LISTTREE]);
                return(usage(myname));
            }
            mbsToUCS2(&monitor_set, argv[0], strlen(argv[0]));
            mbsToUCS2(&monitor_name, argv[1], strlen(argv[1]));
            break;
        case CHECK:
            if (argc < 3)
            {
                printf("Not enough parameters\n");
                return(usage(myname));
            }
            mbsToUCS2(&monitor_set, argv[0], strlen(argv[0]));
            mbsToUCS2(&monitor_name, argv[1], strlen(argv[1]));
            mbsToUCS2(&match, argv[2], strlen(argv[2]));
            break;
    }

    build_param(&connstr, &sap_options);


    /* convert to UCS-2 */
    mbsToUCS2(&buffer, connstr, strlen(connstr));
    mbsToUCS2(&user, sap_options.user, strlen(sap_options.user));

    handle = RfcOpenEx(buffer, &error_info);

    if (handle == RFC_HANDLE_NULL)
    {
        printf("Error: ");
        printfU (cU("(%s) %s\n"), error_info.key, error_info.message);

        return 3;
    }

    switch(mode)
    {
        case SAPINFO:
            get_info(handle);
            break;
        case LISTSETS:
            xmi_logon(handle);
            xmi_getmslist(handle, user);
            xmi_logoff(handle);
            break;
        case LISTMONITORS:
            xmi_logon(handle);
            xmi_getmonlist(handle, user, monitor_set);
            xmi_logoff(handle);
            break;
        case LISTTREE:
            xmi_logon(handle);
            tree_items = xmi_getmontree(handle, user, monitor_set, monitor_name, &tree);
            displaytree(tree_items, tree);
            free(tree);
            xmi_logoff(handle);
            break;
        case CHECK:
            xmi_logon(handle);

            /* TODO read the tree and values out of a cache */
#ifdef CACHE
            if (!(tree_items = tree_from_cache(monitor_set, monitor_name, cache, &tree)))
            {
#endif /* CACHE */
                tree_items = xmi_getmontree(handle, user, monitor_set, monitor_name, &tree);
#ifdef CACHE
                tree_to_cache(monitor_set, monitor_name, tree_items, &tree);
            }
#endif /* CACHE */

            count = matchitems(tree_items, tree, &mteptr, match);
            if (count > 0)
            {
#ifdef DEBUG
                printf("%d matches\n", count);
#endif /* DEBUG */
                mteptr2 = mteptr;
                for (i = 0; i < count; i++)
                {
                    memset(&current, 0, sizeof(current));
                    STRCPY2(current.Mtuid, mteptr2->tid.Mtuid);
                    STRCPY2(current.Mtsysid, mteptr2->tid.Mtsysid);
                    STRCPY2(current.Mtnumrange, mteptr2->tid.Mtnumrange);
                    STRCPY2(current.Mtclass, mteptr2->tid.Mtclass);
                    STRCPY2(current.Mtmcname, mteptr2->tid.Mtmcname);
                    xmi_get_value(handle, user, current, &data);

#ifdef DEBUG
                    printf("  DBG i=%d,val=%d,last=%d  ",i, data.Value, perfval.Lastalstat);
#endif /* DEBUG */
                    /* ignore items without a status */
                    if (data.Value != 0)
                    {

                        if (output_lines++ == 1)
                        {
                            /* we have more than 1 result -> switch to multi-line output mode */

                            /* add the 1st line */
                            STRADD(multi_output, output);
                        }

                        memset(output, 0, sizeof(output));

                        /* add [status] if it is a valid value */
                        if (data.Value> 0 && data.Value < 4) {
                            snprintf(tmpbuf, sizeof(tmpbuf), "[%s] ", exitval[data.Value]);
                            STRADD(output, tmpbuf);
                        }
                        snprintf(tmpbuf, sizeof(tmpbuf), "%s%c", sap_options.dest, PATH_SEPARATOR);
                        STRADD(output, tmpbuf);
                        STRADD(output, mteptr2->name);

#ifdef DEBUG
                        printfU(cU(" Class %s "),mteptr2->tid.Mtclass);
#endif /* DEBUG */
                        /* only print perfdata if the class is 100 */
                        if (!memcmp(mteptr2->tid.Mtclass, MT_CLASS_PERFORMANCE, sizeof(mteptr2->tid.Mtclass)))
                        {
                            xmi_get_perfdata(handle, user, current, &perfval);
                            xmi_get_perfprop(handle, user, current, &perfprop);

#ifdef DEBUG
                            snprintf(tmpbuf, sizeof(tmpbuf), ": %d(warn %d ~ %d, crit %d ~ %d", perfval.Lastperval, perfprop.Treshg2y, perfprop.Treshy2g, perfprop.Treshy2r, perfprop.Treshr2y);
#else
                            snprintf(tmpbuf, sizeof(tmpbuf), ": %d", perfval.Lastperval);
#endif /* DEBUG */
                            STRADD(output, tmpbuf);

                            tmpstr = mystrclean(perfprop.Valunit, 4);
                            UCS2ToMbs(&valunit, tmpstr, strlenU(tmpstr)); /* valunit is defined as RFC_CHAR length 4 */
                            FREE(tmpstr);

                            if (valunit)
                            {
                                snprintf(tmpbuf, sizeof(tmpbuf), " %s", valunit);
                                STRADD(output, tmpbuf);
                            }

                            /* SAP uses thresholds for state changes, but we should assume that the grapher cannot cope
                             * with ranges, so we extract the warning and critical thresholds using the following logic:
                             *
                             *  In status OK:    Warn=green->yellow, Crit=yellow->red
                             *  In status WARN:  Warn=yellow->green, Crit=yellow->red
                             *  In status CRIT:  Warn=yellow->green, Crit=red->yellow
                             *
                             * But if it can then use the '@x:y' format, which is also not correct but brings us closer...
                             */

                            switch(data.Value)
                            {
                                case 1:
                                    warn_threshold = perfprop.Treshg2y;
                                    crit_threshold = perfprop.Treshy2r;
                                    break;
                                case 2:
                                    warn_threshold = perfprop.Treshy2g;
                                    crit_threshold = perfprop.Treshy2r;
                                    break;
                                case 3:
                                    warn_threshold = perfprop.Treshy2g;
                                    crit_threshold = perfprop.Treshr2y;
                                    break;
                            }

                            /* store any perfdata for the grapher */
                            /* first convert the name to a grapher friendly tag */
                            valstr = malloc(strlen(mteptr2->name) + 1);
                            strcpy_nospaces(valstr, mteptr2->name);
                            if (flag_ranges)
                            {
                                snprintf(tmpbuf, sizeof(tmpbuf), "%s=%d%s;@%d:%d;@%d:%d ", valstr, perfval.Lastperval, valunit?valunit:"",
                                    perfprop.Treshy2g, perfprop.Treshg2y, perfprop.Treshr2y, perfprop.Treshy2r);
                            } else if (flag_thresholds) {
                                snprintf(tmpbuf, sizeof(tmpbuf), "%s=%d%s;%d;%d ", valstr, perfval.Lastperval, valunit?valunit:"", warn_threshold, crit_threshold);
                            } else {
                                snprintf(tmpbuf, sizeof(tmpbuf), "%s=%d%s ", valstr, perfval.Lastperval, valunit?valunit:"");
                            }
                            FREE(valunit);
                            FREE(valstr);
                            STRADD(perfdata, tmpbuf);

                        }
                        else if (!memcmp(mteptr2->tid.Mtclass, MT_CLASS_SINGLE_MSG, sizeof(mteptr2->tid.Mtclass)))
                        {
                            xmi_get_status_msg(handle, user, current, &status);

                            tmpstr = mystrclean(status.Msg, sizeofU(status.Msg));
                            UCS2ToMbs(&valstr, tmpstr, strlenU(tmpstr));
                            /* printfU(cU(": %s"), tmpstr); */
                            FREE(tmpstr);

                            snprintf(tmpbuf, sizeof(tmpbuf), ": %s", valstr);
                            FREE(valstr);
                            STRADD(output, tmpbuf);
                        }

                        status_count[data.Value]++;
                        if (data.Value > exitcode) { exitcode = data.Value; }

                        if (output_lines > 1)
                        {

                            STRADD(multi_output, "\n");
                            STRADD(multi_output, output);


                            /* use the first line as a status line */
                            if (status_count[3])
                            {
                                snprintf(output, sizeof(output), "[%s] %d criticals and %d warnings", exitval[exitcode], status_count[3], status_count[2]);
                            } else if (status_count[2]) {
                                snprintf(output, sizeof(output), "[%s] %d warnings", exitval[exitcode], status_count[2]);
                            } else {
                                snprintf(output, sizeof(output), "[%s] %d matching lines found", exitval[exitcode], output_lines);
                            }
                        }
                    }

                    mteptr2++;
                }
            }

            /* now output everything */
            printf("%s", output);
            if (strlen(perfdata)>0) printf(" | %s", perfdata);
            if (strlen(multi_output)>0) printf("\n%s", multi_output);
            printf("\n");

            // free(tids);
            FREE(tree);
            xmi_logoff(handle);
            break;
            
    }


    RfcClose(handle);

    if (exitcode > 0 && exitcode < 4)
    {
        return (exitcode - 1);
    } else {
        return 3;
    }
}

/* this function calls mbstowcs to convert a string to a 4 byte wide-character string (UCS-4)
 * and then packs it into UCS-2 (2 byte)
 *
 * This is a workaround for the gcc option -fshort-wchar
 * because mbstowcs is probably compiled with long wchar
 *
 * Assumptions:
 *  mbctowcs function compiled with 4-byte wchar
 *  this program compiled with 2-byte wchar
 *  no code points exist outside UCS-2 range
 *
 * Ensure that len includes terminating \0 (i.e. strlen+1)
 */
int mbsToUCS2(wchar_t **dest, const char *src, size_t len)
{
    /* worst case memory requirement is 4*len */

    unsigned char   *buffer, *ucs2;
    unsigned char   *from, *to;
    size_t      ret;
    int     i;

    
    buffer = (unsigned char *) malloc((len+1)*4);
    if (buffer == NULL) return -1;


    /* clear buffer to ensure terminating 0 */
    memset(buffer, 0, (len+1)*4);

    ret = mbstowcs((wchar_t *)buffer, src, len);
    if (ret == -1) return -1;
    
#ifdef WCHAR_WORKAROUND
    /* ucs2 should always contain terminated string */
    ucs2 = (unsigned char*) malloc((ret+1)*2);
    memset(ucs2, 0, (ret+1)*2);

    from = buffer;
    to = ucs2;

    for (i=0; *from; i+=4)
    {
        *to = *from;
        *(to+1) = *(from+1);
        from+=4;
        to+=2;
    }

    *dest = (wchar_t *)ucs2;
    free(buffer);
#else
    *dest = (wchar_t *)buffer
#endif /* WCHAR_WORKAROUND */

    return ret;
}
    
/* this function packs a UCS-2 or UCS-4 string into a locale dependent character string (e.g. UTF-8)
 *
 * This is a workaround to allow us to use glibc functions compiled with long wchar
 *
 * Assumptions:
 *  wcstombs function compiled with 4-byte wchar
 *  this program compiled with 2-byte wchar
 *  no code points exist outside UCS-2 range
 *
 * Adds a terminating \0 if not present
 */
int UCS2ToMbs(char **dest, const wchar_t *src, size_t len)
{
    unsigned char   *buffer, *ucs4;
    unsigned char   *from, *to;
    size_t      ret, maxsize;
    int     i;
    wchar_t     *ptr;

    
    maxsize = (len+1)*4;
#ifdef WCHAR_WORKAROUND
    ucs4 = (unsigned char *) malloc(maxsize);
    if (ucs4 == NULL) return -1;


    /* clear ucs4 to ensure terminating 0 */
    memset(ucs4, 0, maxsize);
    to = ucs4;
    from = (unsigned char *)src;

    for (i=0; *from; i+=4)
    {
        *to = *from;
        *(to+1) = *(from+1);
        from+=2;
        to+=4;
    }
#else
    ucs4 = src;
#endif /* WCHAR_WORKAROUND */
    
    /* how much space does the buffer need?? */
    buffer =  (unsigned char *) malloc(maxsize);
    memset(buffer, 0, maxsize);
    ret = wcstombs(buffer, (wchar_t *)ucs4, maxsize);
    if (ret == -1)
    {
        printf("Failed to convert from UCS-2 - check system locale\n");
        exit(3);
    }

    *dest = (char *)realloc(buffer, ret);

    FREE(ucs4);

    return 0;
}
    
    

    
/* call RFC_SYSTEM_INFO */
int get_info(RFC_HANDLE handle)
{
    RFC_RC  rfc_rc = RFC_OK;
    RFC_PARAMETER   exporting[32];
    RFC_PARAMETER   importing[32];
    RFC_TABLE   tables[32];
    RFCSI       rfcsi;
    rfc_char_t  exception[64];
    rfc_char_t  *exceptionp = exception;

    SAP_UC function[] = cU("RFC_SYSTEM_INFO");
    SAP_UC imp[] = cU("RFCSI_EXPORT");

    memset(exporting, 0, sizeof(exporting));
    memset(importing, 0, sizeof(importing));
    memset(tables, 0, sizeof(tables));


    importing[0].name = imp;
    importing[0].nlen = strlenU(imp);
    importing[0].addr = &rfcsi;
    importing[0].leng = sizeofR(rfcsi);
    importing[0].type = RFCTYPE_CHAR;

    
    rfc_rc = RfcCallReceive(handle, function, exporting, importing, tables, &exceptionp);

    if (rfc_rc != RFC_OK)
    {
        printf("Error in %s\n", function);
        exit(3);
    }

    printfU(cU("Connected to SAP release %.*s\n"), sizeofU(rfcsi.rfcsaprl), rfcsi.rfcsaprl);


    
}

void str_param(RFC_PARAMETER *param, SAP_UC *field, SAP_UC *value)
{

    param->name = field;
    param->nlen = strlenU(field);
    param->addr = value;
    param->leng = strlenU(value)*sizeof(SAP_UC);
    param->type = RFCTYPE_CHAR;

}

/* Logon */
int xmi_logon(RFC_HANDLE handle)
{
    RFC_RC  rfc_rc = RFC_OK;
    RFC_PARAMETER   exporting[32];
    RFC_PARAMETER   importing[32];
    RFC_TABLE   tables[32];
    rfc_char_t  exception[64];
    rfc_char_t  *exceptionp = exception;
    int     i;

    SAP_UC function[] = cU("BAPI_XMI_LOGON");
    SAP_UC *exp[] = { cU("EXTCOMPANY"), cU("EXTPRODUCT"), cU("INTERFACE"), cU("VERSION"), NULL};
#ifdef PRETEND_TO_BE_CHECK_SAP
    SAP_UC *exp_val[] = { cU("SAP_MONI_LIB"), cU("LL_MONITORING"), cU("XAL"), cU("1.0"), NULL};
#else
    SAP_UC *exp_val[] = { cU("NETWAYS"), cU("CHECK_CCMS"), cU("XAL"), cU("1.0"), NULL};
#endif /* PRETEND_TO_BE_CHECK_SAP */
    SAP_UC *imp = cU("SESSIONID");
    SAP_UC session_id[25];

    memset(exporting, 0, sizeof(exporting));
    memset(importing, 0, sizeof(importing));
    memset(tables, 0, sizeof(tables));

    for (i=0; exp[i]; i++)
    {
        str_param(&exporting[i], exp[i], exp_val[i]);
    }

    importing[0].name = imp;
    importing[0].nlen = strlenU(imp);
    importing[0].addr = &session_id;
    importing[0].leng = sizeofR(session_id) - 1;
    importing[0].type = RFCTYPE_CHAR;



    rfc_rc = RfcCallReceive(handle, function, exporting, importing, tables, &exceptionp);

    if (rfc_rc != RFC_OK)
    {
        printf("Error in %s\n", function);
        exit(3);
    }

    session_id[sizeofU(session_id) - 1] = '\0';

#ifdef DEBUG
    printfU(cU("Got session ID %s\n"), session_id);
#endif /* DEBUG */

    
}

/* get list of all monitor sets */
int xmi_getmslist(RFC_HANDLE handle, rfc_char_t *user)
{
    RFC_RC  rfc_rc = RFC_OK;
    RFC_PARAMETER   exporting[32];
    RFC_PARAMETER   importing[32];
    RFC_TABLE   tables[32];
    rfc_char_t  exception[64];
    rfc_char_t  *exceptionp = exception;
    rfc_char_t  *tmpstr;
    int     i;

    SAP_UC function[] = cU("BAPI_SYSTEM_MS_GETLIST");
    SAP_UC *exp[] = { cU("EXTERNAL_USER_NAME"), NULL};
    SAP_UC *exp_val[] = { user, NULL};
    SAP_UC *tab = cU("MONITOR_SETS");
    ITAB_H      monitor_sets;
    RFC_TYPEHANDLE  typeHandle;
    BAPIMSNAME  msname;

    memset(exporting, 0, sizeof(exporting));
    memset(importing, 0, sizeof(importing));
    memset(tables, 0, sizeof(tables));

    for (i=0; exp[i]; i++)
    {
        str_param(&exporting[i], exp[i], exp_val[i]);
    }


    rfc_rc = RfcInstallUnicodeStructure( cU("BAPIMSNAME"),
        Description_U_BAPIMSNAME,
        ENTRIES_OF_U_BAPIMSNAME,
        0,
        NULL,
        &typeHandle );

    monitor_sets = ItCreateU( tab, sizeofR(BAPIMSNAME), 0, 0); 

    tables[0].name = tab;
    tables[0].nlen = strlenU(tab);
    tables[0].leng = sizeofR(BAPIMSNAME);
    tables[0].type = typeHandle;
    tables[0].itmode = RFC_ITMODE_BYREFERENCE;
    tables[0].ithandle = monitor_sets;

    rfc_rc = RfcCallReceive(handle, function, exporting, importing, tables, &exceptionp);

    if (rfc_rc != RFC_OK)
    {
        printf("Error in %s\n", function);
        exit(3);
    }

#ifdef DEBUG
    printf("got %d Monitor sets\n", ItFill(monitor_sets));  
#endif /* DEBUG */
    for (i = ItFill(monitor_sets); i; i--)
    {
        ItCopyRow(monitor_sets, i, &msname, 0);
        tmpstr = mystrclean(msname.Name, sizeofU(msname.Name));
        printfU(cU("%s\n"), tmpstr);
        free(tmpstr);
    }
    ItDelete(monitor_sets);

    
}

/* get list of monitors in a set */
int xmi_getmonlist(RFC_HANDLE handle, rfc_char_t *user, rfc_char_t *set)
{
    RFC_RC  rfc_rc = RFC_OK;
    RFC_PARAMETER   exporting[32];
    RFC_PARAMETER   importing[32];
    RFC_TABLE   tables[32];
    rfc_char_t  exception[64];
    rfc_char_t  *exceptionp = exception;
    rfc_char_t  *tmpstr, *tmpstr2;
    int     i;

    SAP_UC function[] = cU("BAPI_SYSTEM_MON_GETLIST");
    SAP_UC *exp[] = { cU("MONI_SET_NAME"), cU("EXTERNAL_USER_NAME"), NULL};
    SAP_UC *exp_val[] = { set, user, NULL};
    SAP_UC *tab = cU("MONITOR_NAMES");
    ITAB_H      monitor_names;
    RFC_TYPEHANDLE  typeHandle;
    BAPIMNNAME  mnname;

    memset(exporting, 0, sizeof(exporting));
    memset(importing, 0, sizeof(importing));
    memset(tables, 0, sizeof(tables));

    for (i=0; exp[i]; i++)
    {
        str_param(&exporting[i], exp[i], exp_val[i]);
    }

    rfc_rc = RfcInstallUnicodeStructure( cU("BAPIMNNAME"),
        Description_U_BAPIMNNAME,
        ENTRIES_OF_U_BAPIMNNAME,
        0,
        NULL,
        &typeHandle );

    monitor_names = ItCreateU( tab, sizeofR(BAPIMNNAME), 0, 0); 

    tables[0].name = tab;
    tables[0].nlen = strlenU(tab);
    tables[0].leng = sizeofR(BAPIMNNAME);
    tables[0].type = typeHandle;
    tables[0].itmode = RFC_ITMODE_BYREFERENCE;
    tables[0].ithandle = monitor_names;


    rfc_rc = RfcCallReceive(handle, function, exporting, importing, tables, &exceptionp);

    if (rfc_rc != RFC_OK)
    {
        printf("Error in %s\n", function);
        exit(3);
    }

#ifdef DEBUG
    printf("got %d Monitor names\n", ItFill(monitor_names));
#endif /* DEBUG */
    for (i = ItFill(monitor_names); i; i--)
    {

        ItCopyRow(monitor_names, i, &mnname, 0);
        tmpstr = mystrclean(mnname.Moni_name, sizeofU(mnname.Moni_name));
        tmpstr2 = mystrclean(mnname.Ms_name, sizeofU(mnname.Ms_name));
        printfU(cU("%s / %s\n"), tmpstr2, tmpstr);
        free(tmpstr);
        free(tmpstr2);
    }
    ItDelete(monitor_names);

    
}


/* get name and parent of an item in the tree */
int get_ancestor(int tree_items, BAPITNDEXT *tree, int idx, char **name)
{
    int i;
    char *name_l;
    wchar_t *cleanstr;

    /* assume that the index is contiguous and starts from 1 */
    
    cleanstr = mystrclean(tree[idx-1].Mtnameshrt, sizeofU(tree[idx-1].Mtnameshrt));
    UCS2ToMbs(&name_l, cleanstr, strlenU(cleanstr));
    
    asprintf(&name_l, "%s%c%s", name_l, PATH_SEPARATOR, *name);
    FREE(*name);
    *name = name_l;
    return tree[idx-1].Alparintre;
    
}

/* try matching a regex against a tree
 *
 * Returns a list of TIDs to allow values to be queried
 */
int matchitems(int tree_items, BAPITNDEXT *tree, MTE **mteptrs, wchar_t *match2)
{
    char        *match, *test;
    wchar_t     *cleanstr;
    int         i, status, matchcount = 0;
    regex_t     re;
    BAPITID     tid;
    MTE     *mte;
    int     parent_id;
    char        *full_name = 0;
    BAPIPACVAL      perfval;



    /* in order to use external regex functions we need to convert the strings
     * back to locale - will break if our locale doesn't support a character...
     */

    UCS2ToMbs(&match, match2, strlenU(match2));


    /* How many MTE pointers do we need to return?
     * Assume that every line matches
     */
    mte = malloc(sizeof(struct MTE) * tree_items+10);
    if(!mte) printf("failed to allocate memory\n");
    *mteptrs = mte;


    status = regcomp(&re, match, REG_ICASE|REG_EXTENDED|REG_NOSUB);
    if (status != 0) { printf("Error compiling regex\n"); exit(3); }

    /* loop over the array and look for a match
     */

    for (i = 0; i < tree_items; i++)
    {
        cleanstr = mystrclean(tree[i].Mtnameshrt, sizeofU(tree[i].Mtnameshrt));

        UCS2ToMbs(&test, cleanstr, strlenU(cleanstr));

        if (flag_fullpath)
        {
            if (tree[i].Alparintre > 0)
            {
#ifdef DEBUG
                // printf("Attempting to build full path to MTE\n");
#endif /* DEBUG */
                parent_id =  tree[i].Alparintre;
                asprintf(&full_name, "%s", test);
                while (parent_id = get_ancestor(tree_items, tree, parent_id, &full_name))
                {
                }
                FREE(test);
                test = full_name;
                full_name = 0;
            }
        }

        status = regexec(&re, test, (size_t) 0, NULL, 0);
        if (!status)
        {
#ifdef DEBUG
            printf("%d: %s matches our regex\n", i, test);
#endif /* DEBUG */

            memset(mte, 0, sizeof(struct MTE));

            STRCPYUSTR(mte->name_u2, cleanstr);
            STRCPYSTR(mte->name, test);
#ifdef DEBUG
            printf("mte->name is %s\n", mte->name);
#endif /* DEBUG */
            STRCPY(mte->tid.Mtuid, tree[i].Mtuid);
            STRCPY(mte->tid.Mtsysid, tree[i].Mtsysid);
            STRCPY(mte->tid.Mtnumrange, tree[i].Mtnumrange);
            STRCPY(mte->tid.Mtclass, tree[i].Mtclass);
            STRCPY(mte->tid.Mtmcname, tree[i].Mtmcname);
            /* should be able to reference the MTE with its UID */
/*
            STRCPY(mte->tid.Mtindex, tree[i].Mtindex);
            STRCPY(mte->tid.Extindex, tree[i].Extindex);
*/

        mte++;
            matchcount++;

        }
        FREE(cleanstr);
        FREE(test);
    }
    FREE(match);

/*
    // don't bother to free up unused memory
    *mteptrs = realloc(*mteptrs, sizeof(struct MTE) * matchcount);

*/

    return matchcount;
}


/* display a monitor tree
 */
int displaytree(int tree_items, BAPITNDEXT *tree)
{
    int     i, parent = 0;
    BAPITNDEXT  *tree_item;


    /* loop over the array and ASSUME that the tree is sorted
     * may not be true, but it saves time...
     */

    for (i = 0; i < tree_items; i++)
    {

#ifdef DEBUG
        printfU(cU("[%d]"), tree[i].Alidxintre);
#endif /* DEBUG */
        printfU(cU("%*s%.*s"), tree[i].Allevintre * 2, cU(""), sizeofU(tree[i].Mtnameshrt), tree[i].Mtnameshrt);
#ifdef DEBUG
        printfU(cU("\t\t(child of %d, class is %.*s)"),  tree[i].Alparintre, sizeofU(tree[i].Mtclass), tree[i].Mtclass);
#endif /* DEBUG */
        printfU(cU("\n"));

    }
}


/* get monitor tree
 * fills an array of tree items (BAPITNDEXT)
 * which includes the ID structure necessary to read values
 */
int xmi_getmontree(RFC_HANDLE handle, rfc_char_t *user, rfc_char_t *set, rfc_char_t *mon, BAPITNDEXT **tree)
{
    RFC_RC  rfc_rc = RFC_OK;
    RFC_PARAMETER   exporting[32];
    RFC_PARAMETER   importing[32];
    RFC_TABLE   tables[32];
    rfc_char_t  exception[64];
    rfc_char_t  *exceptionp = exception;
    int     i;

    SAP_UC function[] = cU("BAPI_SYSTEM_MON_GETTREE");
    SAP_UC *exp[] = { cU("MONITOR_NAME"), cU("EXTERNAL_USER_NAME"), NULL};
    SAP_UC *tab = cU("TREE_NODES");
    ITAB_H      monitor_tree;
    RFC_TYPEHANDLE  typeHandle, typeHandle2;
    BAPIMNNAME  monitor_name;
    BAPITNDEXT  *tree_item;
    BAPITID     item_tid;
    int     number_of_items;


#ifdef DEBUG
    printfU(cU("%s / %s\n"), set, mon);
#endif /* DEBUG */

    /* fill the BAPIMNNAME structure with the set and monitor name */
    mystrfill(monitor_name.Moni_name, ' ', sizeofU(monitor_name.Moni_name));
    mystrfill(monitor_name.Ms_name, ' ', sizeofU(monitor_name.Ms_name));
    memcpy(monitor_name.Moni_name, mon, (strlenU(mon))*2);
    memcpy(monitor_name.Ms_name, set, (strlenU(set))*2);

    memset(exporting, 0, sizeof(exporting));
    memset(importing, 0, sizeof(importing));
    memset(tables, 0, sizeof(tables));

    rfc_rc = RfcInstallUnicodeStructure( cU("BAPIMNNAME"),
        Description_U_BAPIMNNAME,
        ENTRIES_OF_U_BAPIMNNAME,
        0,
        NULL,
        &typeHandle );

    exporting[0].name = exp[0];
    exporting[0].nlen = strlenU(exp[0]);
    exporting[0].addr = &monitor_name;
    exporting[0].leng = sizeofR(BAPIMNNAME);
    exporting[0].type = typeHandle;

    str_param(&exporting[1], exp[1], user);

    rfc_rc = RfcInstallUnicodeStructure( cU("BAPITNDEXT"),
        Description_U_BAPITNDEXT,
        ENTRIES_OF_U_BAPITNDEXT,
        0,
        NULL,
        &typeHandle2 );

    monitor_tree = ItCreateU( tab, sizeofR(BAPITNDEXT), 0, 0); 

    tables[0].name = tab;
    tables[0].nlen = strlenU(tab);
    tables[0].leng = sizeofR(BAPITNDEXT);
    tables[0].type = typeHandle2;
    tables[0].itmode = RFC_ITMODE_BYREFERENCE;
    tables[0].ithandle = monitor_tree;


    rfc_rc = RfcCallReceive(handle, function, exporting, importing, tables, &exceptionp);

    if (rfc_rc != RFC_OK)
    {
        printf("Error in %s\n", function);
        exit(3);
    }

    number_of_items = ItFill(monitor_tree);
#ifdef DEBUG
    printf("got %d tree items\n", number_of_items);
#endif /* DEBUG */

    *tree = malloc(sizeofR(struct BAPITNDEXT) * number_of_items);
    memset(*tree, 0, sizeofR(struct BAPITNDEXT) * number_of_items);
    tree_item = *tree;
    for (i = 1; i<=number_of_items; i++)
    {
        ItCopyRow(monitor_tree, i, tree_item++, 0);
    }
    ItDelete(monitor_tree);
    return number_of_items;

    
}

#ifdef CACHE
int tree_from_cache(rfc_char_t *set, rfc_char_t *mon, int cache, BAPITNDEXT **tree)
{
    /*
     * create cache filename (md5sum?)
     * stat file & check age / size for plausibility
     * open with timeout
     * calc number_of_items from st.st_size / sizeofR(struct BAPITNDEXT)
     * mmap to tree var
     */
    int number_of_items = 0;
    return number_of_items;
}

int tree_to_cache(rfc_char_t *set, rfc_char_t *mon, int number_of_items, BAPITNDEXT **tree)
{
    /*
     * refuse to create cache as root
     * create cache filename (md5sum?)
     * warn if another user owns the cache
     * exclusively lock file / with timeout
     * seek to sizeofR(struct BAPITNDEXT) * number_of_items
     * mmap
     * memcpy
     */
}
#endif /* CACHE */

int xmi_get_perfdata(RFC_HANDLE handle, rfc_char_t *user, BAPITID item_tid, BAPIPACVAL *retstr)
{
    RFC_RC  rfc_rc = RFC_OK;
    RFC_PARAMETER   exporting[32];
    RFC_PARAMETER   importing[32];
    RFC_TABLE   tables[32];
    rfc_char_t  exception[64];
    rfc_char_t  *exceptionp = exception;
    int     i;

    SAP_UC function[] = cU("BAPI_SYSTEM_MTE_GETPERFCURVAL");
    SAP_UC exptid[] =  cU("TID");
    SAP_UC *exp[] = { cU("EXTERNAL_USER_NAME"), NULL};
    SAP_UC *exp_val[] = { user, NULL};
    SAP_UC imp[] =  cU("CURRENT_VALUE");
    RFC_TYPEHANDLE  typeHandle, typeHandle2;
    BAPIPACVAL  values;

    memset(exporting, 0, sizeof(exporting));
    memset(importing, 0, sizeof(importing));
    memset(tables, 0, sizeof(tables));
    memset(&values, 0, sizeofR(values));

    for (i=0; exp[i]; i++)
    {
        str_param(&exporting[i], exp[i], exp_val[i]);
    }

    rfc_rc = RfcInstallUnicodeStructure( cU("BAPITID"),
        Description_U_BAPITID,
        ENTRIES_OF_U_BAPITID,
        0,
        NULL,
        &typeHandle );

    exporting[1].name = exptid;
    exporting[1].nlen = strlenU(exptid);
    exporting[1].addr = &item_tid;
    exporting[1].leng = sizeofR(BAPITID);
    exporting[1].type = typeHandle;


    rfc_rc = RfcInstallUnicodeStructure( cU("BAPIPACVAL"),
        Description_U_BAPIPACVAL,
        ENTRIES_OF_U_BAPIPACVAL,
        0,
        NULL,
        &typeHandle2 );

    importing[0].name = imp;
    importing[0].nlen = strlenU(imp);
    importing[0].leng = sizeofR(BAPIPACVAL);
    importing[0].type = typeHandle2;
    importing[0].addr = &values;


    rfc_rc = RfcCallReceive(handle, function, exporting, importing, tables, &exceptionp);

    if (rfc_rc != RFC_OK)
    {
        printf("Error in %s\n", function);
        exit(3);
    }

#ifdef DEBUG2
    printfU(cU("  Got value Alrelevval %d\n"), values.Alrelevval);
    printfU(cU("  Got value Alrelvaldt %.*s\n"), sizeofU(values.Alrelvaldt), values.Alrelvaldt);
    printfU(cU("  Got value Alrelvalti %.*s\n"), sizeofU(values.Alrelvalti), values.Alrelvalti);
    printfU(cU("  Got value Lastalstat %d\n"), values.Lastalstat);
    printfU(cU("  Got value Lastperval %d\n"), values.Lastperval);
    printfU(cU("  Got value Avg01pval %d\n"), values.Avg01pval);
    printfU(cU("  Got value Avg05pval %d\n"), values.Avg05pval);
    printfU(cU("  Got value Avg15pval %d\n"), values.Avg15pval);
    printfU(cU("  Got value Avg01sval %d\n"), values.Avg01sval);
    printfU(cU("  Got value Avg05sval %d\n"), values.Avg05sval);
    printfU(cU("  Got value Avg15sval %d\n"), values.Avg15sval);
    printfU(cU("  Got value Avg01cval %d\n"), values.Avg01cval);
    printfU(cU("  Got value Avg05cval %d\n"), values.Avg05cval);
    printfU(cU("  Got value Avg15cval %d\n"), values.Avg15cval);
    printfU(cU("  Got value Maxpfvalue %d\n"), values.Maxpfvalue);
    printfU(cU("  Got value Maxpfdate %.*s\n"), sizeofU(values.Maxpfdate), values.Maxpfdate);
    printfU(cU("  Got value Maxpftime %.*s\n"), sizeofU(values.Maxpftime), values.Maxpftime);
    printfU(cU("  Got value Minpfvalue %d\n"), values.Minpfvalue);
    printfU(cU("  Got value Minpfdate %.*s\n"), sizeofU(values.Minpfdate), values.Minpfdate);
    printfU(cU("  Got value Minpftime %.*s\n"), sizeofU(values.Minpftime), values.Minpftime);
#endif /* DEBUG */

    *retstr = values;

}


/* Get properties of perf data */
int xmi_get_perfprop(RFC_HANDLE handle, rfc_char_t *user, BAPITID item_tid, BAPIPAPROP *retstr)
{
    RFC_RC  rfc_rc = RFC_OK;
    RFC_PARAMETER   exporting[32];
    RFC_PARAMETER   importing[32];
    RFC_TABLE   tables[32];
    rfc_char_t  exception[64];
    rfc_char_t  *exceptionp = exception;
    int     i;

    SAP_UC function[] = cU("BAPI_SYSTEM_MTE_GETPERFPROP");
    SAP_UC exptid[] =  cU("TID");
    SAP_UC *exp[] = { cU("EXTERNAL_USER_NAME"), NULL};
    SAP_UC *exp_val[] = { user, NULL};
    SAP_UC imp[] =  cU("PROPERTIES");
    RFC_TYPEHANDLE  typeHandle, typeHandle2;
    BAPIPAPROP  values;

    memset(exporting, 0, sizeof(exporting));
    memset(importing, 0, sizeof(importing));
    memset(tables, 0, sizeof(tables));
    memset(&values, 0, sizeofR(values));

    for (i=0; exp[i]; i++)
    {
        str_param(&exporting[i], exp[i], exp_val[i]);
    }

    rfc_rc = RfcInstallUnicodeStructure( cU("BAPITID"),
        Description_U_BAPITID,
        ENTRIES_OF_U_BAPITID,
        0,
        NULL,
        &typeHandle );

    exporting[1].name = exptid;
    exporting[1].nlen = strlenU(exptid);
    exporting[1].addr = &item_tid;
    exporting[1].leng = sizeofR(BAPITID);
    exporting[1].type = typeHandle;


    rfc_rc = RfcInstallUnicodeStructure( cU("BAPIPAPROP"),
        Description_U_BAPIPAPROP,
        ENTRIES_OF_U_BAPIPAPROP,
        0,
        NULL,
        &typeHandle2 );

    importing[0].name = imp;
    importing[0].nlen = strlenU(imp);
    importing[0].leng = sizeofR(BAPIPAPROP);
    importing[0].type = typeHandle2;
    importing[0].addr = &values;


    rfc_rc = RfcCallReceive(handle, function, exporting, importing, tables, &exceptionp);

    if (rfc_rc != RFC_OK)
    {
        printf("Error in %s\n", function);
        exit(3);
    }

    *retstr = values;

}

/* Get last status message */
int xmi_get_status_msg(RFC_HANDLE handle, rfc_char_t *user, BAPITID item_tid, BAPISMCVAL *retstr)
{
    RFC_RC  rfc_rc = RFC_OK;
    RFC_PARAMETER   exporting[32];
    RFC_PARAMETER   importing[32];
    RFC_TABLE   tables[32];
    rfc_char_t  exception[64];
    rfc_char_t  *exceptionp = exception;
    int     i;

    SAP_UC function[] = cU("BAPI_SYSTEM_MTE_GETSMVALUE");
    SAP_UC exptid[] =  cU("TID");
    SAP_UC *exp[] = { cU("EXTERNAL_USER_NAME"), NULL};
    SAP_UC *exp_val[] = { user, NULL};
    SAP_UC imp[] =  cU("VALUE");
    RFC_TYPEHANDLE  typeHandle, typeHandle2;
    BAPISMCVAL  values;

    memset(exporting, 0, sizeof(exporting));
    memset(importing, 0, sizeof(importing));
    memset(tables, 0, sizeof(tables));
    memset(&values, 0, sizeofR(values));

    for (i=0; exp[i]; i++)
    {
        str_param(&exporting[i], exp[i], exp_val[i]);
    }

    rfc_rc = RfcInstallUnicodeStructure( cU("BAPITID"),
        Description_U_BAPITID,
        ENTRIES_OF_U_BAPITID,
        0,
        NULL,
        &typeHandle );

    exporting[1].name = exptid;
    exporting[1].nlen = strlenU(exptid);
    exporting[1].addr = &item_tid;
    exporting[1].leng = sizeofR(BAPITID);
    exporting[1].type = typeHandle;


    rfc_rc = RfcInstallUnicodeStructure( cU("BAPISMCVAL"),
        Description_U_BAPISMCVAL,
        ENTRIES_OF_U_BAPISMCVAL,
        0,
        NULL,
        &typeHandle2 );

    importing[0].name = imp;
    importing[0].nlen = strlenU(imp);
    importing[0].leng = sizeofR(BAPISMCVAL);
    importing[0].type = typeHandle2;
    importing[0].addr = &values;


    rfc_rc = RfcCallReceive(handle, function, exporting, importing, tables, &exceptionp);

    if (rfc_rc != RFC_OK)
    {
        printf("Error in %s\n", function);
        exit(3);
    }

    *retstr = values;

}


/* get general properties */
int xmi_get_value(RFC_HANDLE handle, rfc_char_t *user, BAPITID item_tid, BAPIALDATA *retstr)
{
    RFC_RC  rfc_rc = RFC_OK;
    RFC_PARAMETER   exporting[32];
    RFC_PARAMETER   importing[32];
    RFC_TABLE   tables[32];
    rfc_char_t  exception[64];
    rfc_char_t  *exceptionp = exception;
    int     i;

    SAP_UC function[] = cU("BAPI_SYSTEM_MTE_GETGENPROP");
    SAP_UC exptid[] =  cU("TID");
    SAP_UC *exp[] = { cU("EXTERNAL_USER_NAME"), NULL};
    SAP_UC *exp_val[] = { user, NULL};
    SAP_UC imp[] =  cU("ACTUAL_ALERT_DATA");
    RFC_TYPEHANDLE  typeHandle, typeHandle2;
    BAPIALDATA  values;

    memset(exporting, 0, sizeof(exporting));
    memset(importing, 0, sizeof(importing));
    memset(tables, 0, sizeof(tables));
    memset(&values, 0, sizeofR(values));

    for (i=0; exp[i]; i++)
    {
        str_param(&exporting[i], exp[i], exp_val[i]);
    }

    rfc_rc = RfcInstallUnicodeStructure( cU("BAPITID"),
        Description_U_BAPITID,
        ENTRIES_OF_U_BAPITID,
        0,
        NULL,
        &typeHandle );

    exporting[1].name = exptid;
    exporting[1].nlen = strlenU(exptid);
    exporting[1].addr = &item_tid;
    exporting[1].leng = sizeofR(BAPITID);
    exporting[1].type = typeHandle;


    rfc_rc = RfcInstallUnicodeStructure( cU("BAPIALDATA"),
        Description_U_BAPIALDATA,
        ENTRIES_OF_U_BAPIALDATA,
        0,
        NULL,
        &typeHandle2 );

    importing[0].name = imp;
    importing[0].nlen = strlenU(imp);
    importing[0].leng = sizeofR(BAPIALDATA);
    importing[0].type = typeHandle2;
    importing[0].addr = &values;


    rfc_rc = RfcCallReceive(handle, function, exporting, importing, tables, &exceptionp);

    if (rfc_rc != RFC_OK)
    {
        printf("Error in %s\n", function);
        exit(3);
    }

    *retstr = values;

}
/* Logoff */
int xmi_logoff(RFC_HANDLE handle)
{
    RFC_RC  rfc_rc = RFC_OK;
    RFC_PARAMETER   exporting[32];
    RFC_PARAMETER   importing[32];
    RFC_TABLE   tables[32];
    rfc_char_t  exception[64];
    rfc_char_t  *exceptionp = exception;
    int     i;

    SAP_UC function[] = cU("BAPI_XMI_LOGOFF");
    SAP_UC *exp[] = { cU("INTERFACE"), NULL};
    SAP_UC *exp_val[] = { cU("XAL"), NULL};

    memset(&exporting, 0, sizeof(exporting));
    memset(&importing, 0, sizeof(importing));
    memset(&tables, 0, sizeof(tables));

    for (i=0; exp[i]; i++)
    {
        str_param(&exporting[i], exp[i], exp_val[i]);
    }


    rfc_rc = RfcCallReceive(handle, function, exporting, importing, tables, &exceptionp);

    if (rfc_rc != RFC_OK)
    {
        printf("Error in %s\n", function);
        exit(3);
    }

#ifdef DEBUG
    printfU(cU("Logged off\n"));
#endif /* DEBUG */
    
}

size_t mystrlen(const wchar_t *s)
{
    size_t n = 0;
    while (*s++) n++;
    return(n);
}

size_t mystrfill(wchar_t *s, const wchar_t c, size_t n)
{
    int i;

    for (i=0; i<n; i++)
    {
        (*s++) = c;
    }   
    return(n);
}

/* only use for null-terminated strings we already know the size of */
void strcpy_nospaces(char *dest, char *src)
{
    static unsigned char allowed[256] = "_________________________________!_#_%__()*+,-.-0123456789_____?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[_]^__abcdefghijklmnopqrstuvwxyz{_}__________________________________________________________________________________________________________________________________";

    while(*src)
    {
        *(dest++) = allowed[(unsigned char) *(src++)];
    }
    *dest = 0;
}

/* remove trailing spaces and null terminate a string
 * N.B. returns a new string because we can't be sure
 * that we have enough space for the trailing 0 in the
 * original string
 */
wchar_t *mystrclean(wchar_t *s, size_t n)
{
    int i, endpos = 0;
    wchar_t *newstr;


    for (i=n; i; i--)
    {
        if (!endpos && *(s+i-1) != ' ')
        {
            endpos = i;
        }
    }   
    
    newstr =  malloc(sizeof(wchar_t)*(endpos+1));
    if (newstr == NULL)
    {
        printf("Out of Memory\n");
        exit(3);
    }

    for (i=endpos; i--; )
    {
        *(newstr+i) = *(s+i);
    }
    newstr[endpos] = L'\0';
    return newstr;
}

/* build a SAP connection parameter string from the options */
int build_param(char **param, struct sap_options_s *opt)
{

    char **tmp;
    char svc[8];

    asprintf(param, "TYPE=%d USER=%s PASSWD=%s CLIENT=%s LANG=%s TRACE=%d",
        opt->ver, quote_param(opt->user), quote_param(opt->pass), opt->client, opt->lang, opt->trace);
    
    if (opt->ver > 2)
    {
        asprintf(param, "%s ASHOST=%s SYSNR=%s", *param, opt->host, opt->sysnr);
    } else {
        if (!opt->svc)
        {
            snprintf(svc, sizeof(svc), "sapgw%s", opt->sysnr);
        }
        asprintf(param, "%s DEST=%s GWHOST=%s GWSERV=%s", *param, quote_param(opt->dest), opt->gw, opt->svc?opt->svc:svc);
    }
    return 0;
}

char *quote_param(char *param)
{
    char *tmp;

    tmp = param;

    if (param && strchr(param, ' '))
    {
        asprintf(&tmp, "...%s...", param);
    }
    return tmp;
}

int usage(char *me)
{
    int i;

    printf("Usage: %s <options> <Monitor Set> [<Monitor Name> [<Item>]]\n", me);
    printf("\n where\n");
    printf("  --host|-h\t\tSAP Hostname\n");
    printf("  --user|-u\t\tSAP RFC user\n");
    printf("  --pass|-p\t\tPassword for this user\n");
    printf("  --client|-c\t\tClient (default 000)\n");
    printf("  --sysnr|-s\t\tSystem number (default 00)\n");
    printf("  --lang|-l\t\tLanguage (default E)\n");
    printf("  --trace|-t\t\tCreate .trc file in working directory\n");
    printf("  -2\t\t\tUse Version 2\n");
    printf("  --dest|-d\t\tSAP Destination (V2)\n");
    printf("  --gateway|-g\t\tSAP Gateway (V2)\n");
    printf("  --gwserv|-g\t\tSAP Gateway Server (V2 - defaults to sapgwxx where xx is the sysnr)\n");
    printf("\n");
    printf("  --client|-c\t\tClient (default 000)\n");
    printf("  --sysnr|-s\t\tSystem number (default 00)\n");
    printf("  --lang|-l\t\tLanguage (default E)\n");
    printf("  --trace|-t\t\tCreate .trc file in working directory\n");
    printf("  --fullpath|-f\t\tMatch against the full path of the Object (Separator is %c )\n",
        PATH_SEPARATOR);
    printf("  --thresholds\t\tInclude the SAP thresholds in the perf data (simplified)\n");
    printf("  --use-ranges\t\tInclude the SAP thresholds in the perf data (range format)\n");
#ifdef CACHE
    printf("  --cache\t\tEnable cacheing of the monitor tree for this number of seconds\n");
#endif /* CACHE */
    printf("\n  --mode|-m\t\tBeing one of: ");
    for (i=0; modes[i]; i++)
    {
        printf("%s ", modes[i]);
    }
    printf("\n\n");
    return 3;
}


